import argparse
from collections import namedtuple

import torch
import torchvision
from tqdm import trange

import tvdcn
from utils.time_meter import TimeMeter

TimeMetersPair = namedtuple('ForwardBackwardTimeMetersPair',
                            ['forward', 'backward'])


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--deformable', type=bool, default=True)
    parser.add_argument('--modulated', type=bool, default=True)
    parser.add_argument('--bias', type=bool, default=True)
    parser.add_argument('--device', default='cuda')
    parser.add_argument('-n', '--n', type=int, default=1000)
    return parser.parse_args()


def main():
    args = parse_args()
    device = torch.device(args.device)

    torch.manual_seed(1)
    input = torch.randn(1, 4, 32, 32, device=device, requires_grad=True)
    weight = torch.randn(2, 2, 5, 5, device=device, requires_grad=True)
    offset = torch.randn(1, 50, 28, 13, device=device, requires_grad=True) if args.deformable else None
    mask = torch.randn(1, 25, 28, 13, device=device, requires_grad=True) if args.modulated else None
    bias = torch.randn(2, device=device, requires_grad=True) if args.bias else None

    stride = (1, 2)
    padding = (0, 1)
    dilation = (1, 2)
    groups = 2

    params = [input, weight]
    if args.deformable:
        params.append(offset)
    if args.modulated:
        params.append(mask)
    if args.bias:
        params.append(bias)
    optim = torch.optim.Optimizer(params, {})

    # aliasing
    torchvision_f = torchvision.ops.deform_conv2d
    torchvision_args = (input, offset, weight, bias, stride, padding, dilation, mask)
    tvdcn_f = tvdcn.ops.deform_conv2d
    tvdcn_args = (input, weight, offset, mask, bias, stride, padding, dilation, groups)

    # warmup
    output = torchvision_f(*torchvision_args) + tvdcn_f(*tvdcn_args)
    grad = torch.ones_like(output)
    output.backward(grad)
    optim.zero_grad()

    torchvision_time_meter = TimeMetersPair(TimeMeter(), TimeMeter())
    for iter_id in (pbar := trange(args.n, desc=torchvision_f.__name__)):
        optim.zero_grad()
        with torchvision_time_meter.forward:
            output = torchvision_f(*torchvision_args)
        with torchvision_time_meter.backward:
            output.backward(grad)
        pbar.set_description(f'[{torchvision_f.__module__}.{torchvision_f.__name__} - iter {iter_id + 1}], '
                             f'forward_fps={torchvision_time_meter.forward.fps:.04f}, '
                             f'backward_fps={torchvision_time_meter.backward.fps:.04f}')

    tvdcn_time_meter = TimeMetersPair(TimeMeter(), TimeMeter())
    for iter_id in (pbar := trange(args.n, desc=tvdcn_f.__name__)):
        optim.zero_grad()
        with tvdcn_time_meter.forward:
            output = tvdcn_f(*tvdcn_args)
        with tvdcn_time_meter.backward:
            output.backward(grad)
        pbar.set_description(f'[{tvdcn_f.__module__}.{tvdcn_f.__name__} - iter {iter_id + 1}], '
                             f'forward_fps={tvdcn_time_meter.forward.fps:.04f}, '
                             f'backward_fps={tvdcn_time_meter.backward.fps:.04f}')


if __name__ == '__main__':
    main()



from typing import Tuple, Optional

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.common_types import _size_any_t, _size_1_t, _size_2_t, _size_3_t
from torch.nn.modules.utils import _single, _pair, _triple

__all__ = [
    'mask_softmax1d',
    'mask_softmax2d',
    'mask_softmax3d',
    'MaskSoftmax1d',
    'MaskSoftmax2d',
    'MaskSoftmax3d',
]


def mask_softmax1d(mask: Tensor,
                   kernel_size: Tuple[int]) -> Tensor:
    r"""
    Performs 1D Mask Softmax Normalization.

    Arguments:
        mask (Tensor[batch_size, mask_groups * kernel_width, out_width]): modulation
            masks to be multiplied with each output of convolution kernel.
        kernel_size (int or Tuple[int]): convolution kernel size.
    """
    batch_size, _, out_width = mask.size()

    weight_w, = _single(kernel_size)
    mask_groups = mask.size(1) // weight_w

    if mask_groups == 0:
        raise RuntimeError(
            "The shape of the mask tensor at dimension 1 is not valid. It should "
            "be a multiple of kernel_size[0].\n"
            "Got mask.size(1)={}, while kernel_size[0]={}".format(
                mask.size(1), weight_w))

    mask = mask.view(batch_size,
                     mask_groups,
                     weight_w,
                     out_width)
    mask = torch.softmax(mask, dim=2)
    mask = mask.view(batch_size,
                     mask_groups * weight_w,
                     out_width)
    return mask


def mask_softmax2d(mask: Tensor,
                   kernel_size: Tuple[int, int]) -> Tensor:
    r"""
    Performs 2D Mask Softmax Normalization.

    References:
        https://arxiv.org/abs/2211.05778

    Arguments:
        mask (Tensor[batch_size, mask_groups * kernel_height * kernel_width,
            out_height, out_width]): modulation masks to be multiplied with each output
            of convolution kernel.
        kernel_size (int or Tuple[int, int]): convolution kernel size.
    """
    batch_size, _, out_height, out_width = mask.size()

    weight_h, weight_w = _pair(kernel_size)
    mask_groups = mask.size(1) // (weight_h * weight_w)

    if mask_groups == 0:
        raise RuntimeError(
            "The shape of the mask tensor at dimension 1 is not valid. It should "
            "be a multiple of kernel_size[0] * kernel_size[1].\n"
            "Got mask.size(1)={}, while kernel_size[0] * kernel_size[1]={}".format(
                mask.size(1), weight_h * weight_w))

    mask = mask.view(batch_size,
                     mask_groups,
                     weight_h * weight_w,
                     out_height,
                     out_width)
    mask = torch.softmax(mask, dim=2)
    mask = mask.view(batch_size,
                     mask_groups * weight_h * weight_w,
                     out_height,
                     out_width)
    return mask


def mask_softmax3d(mask: Tensor,
                   kernel_size: Tuple[int, int, int]) -> Tensor:
    r"""
    Performs 3D Mask Softmax Normalization.

    Arguments:
        mask (Tensor[batch_size, mask_groups * kernel_depth * kernel_height * kernel_width,
            out_depth, out_height, out_width]): modulation masks to be multiplied with each output
            of convolution kernel. Default: None
        kernel_size (int or Tuple[int, int, int]): convolution kernel size.
    """
    batch_size, _, out_depth, out_height, out_width = mask.size()

    weight_d, weight_h, weight_w = _triple(kernel_size)
    mask_groups = mask.size(1) // (weight_d * weight_h * weight_w)

    if mask_groups == 0:
        raise RuntimeError(
            "The shape of the mask tensor at dimension 1 is not valid. It should "
            "be a multiple of kernel_size[0] * kernel_size[1] * kernel_size[2].\n"
            "Got mask.size(1)={}, while kernel_size[0] * kernel_size[1] * kernel_size[2]={}".format(
                mask.size(1), weight_d * weight_h * weight_w))

    mask = mask.view(batch_size,
                     mask_groups,
                     weight_d * weight_h * weight_w,
                     out_depth,
                     out_height,
                     out_width)
    mask = torch.softmax(mask, dim=2)
    mask = mask.view(batch_size,
                     mask_groups * weight_d * weight_h * weight_w,
                     out_depth,
                     out_height,
                     out_width)
    return mask


################################################################################
# Modules
################################################################################
class _MaskSoftmaxNd(nn.Module):
    """
    Base class for MaskSoftmax
    """

    def __init__(self, kernel_size: _size_any_t):
        super().__init__()
        self.kernel_size: Tuple[int, ...] = kernel_size

    def forward(self, mask: Tensor, kernel_size: Optional[_size_any_t] = None) -> Tensor:
        raise NotImplementedError


class MaskSoftmax1d(_MaskSoftmaxNd):
    """
    See :func:`mask_softmax1d`
    """

    def __init__(self, kernel_size: _size_1_t) -> None:
        kernel_size = _single(kernel_size)
        super().__init__(kernel_size)

    def forward(self, mask: Tensor, kernel_size: Optional[Tuple[int]] = None) -> Tensor:
        kernel_size = kernel_size if kernel_size is not None else self.kernel_size
        return mask_softmax1d(mask, kernel_size)


class MaskSoftmax2d(_MaskSoftmaxNd):
    """
    See :func:`mask_softmax2d`
    """

    def __init__(self, kernel_size: _size_2_t) -> None:
        kernel_size = _pair(kernel_size)
        super().__init__(kernel_size)

    def forward(self, mask: Tensor, kernel_size: Optional[Tuple[int, int]] = None) -> Tensor:
        kernel_size = kernel_size if kernel_size is not None else self.kernel_size
        return mask_softmax2d(mask, kernel_size)  # type: ignore[arg-type]


class MaskSoftmax3d(_MaskSoftmaxNd):
    """
    See :func:`mask_softmax3d`
    """

    def __init__(self, kernel_size: _size_3_t) -> None:
        kernel_size = _triple(kernel_size)
        super().__init__(kernel_size)

    def forward(self, mask: Tensor, kernel_size: Optional[Tuple[int, int, int]] = None) -> Tensor:
        kernel_size = kernel_size if kernel_size is not None else self.kernel_size
        return mask_softmax3d(mask, kernel_size)  # type: ignore[arg-type]



import os
from types import FunctionType
from typing import Any

import torch

extension_namespace = os.path.basename(os.path.dirname(__file__))


def _log_api_usage_once(obj: Any) -> None:
    """
    Logs API usage(module and name) within an organization.
    In a large ecosystem, it's often useful to track the PyTorch and
    Extension APIs usage. This API provides the similar functionality to the
    logging module in the Python stdlib. It can be used for debugging purpose
    to log which methods are used and by default it is inactive, unless the user
    manually subscribes a logger via the `SetAPIUsageLogger method <https://github.com/pytorch/pytorch/blob/eb3b9fe719b21fae13c7a7cf3253f970290a573e/c10/util/Logging.cpp#L114>`_.
    Please note it is triggered only once for the same API call within a process.
    It does not collect any data from open-source users since it is no-op by default.
    For more information, please refer to
    * PyTorch note: https://pytorch.org/docs/stable/notes/large_scale_deployments.html#api-usage-logging;
    * Logging policy: https://github.com/pytorch/vision/issues/5052;
    Args:
        obj (class instance or method): an object to extract info from.
    """
    module = obj.__module__
    if not module.startswith(extension_namespace):
        module = f'{extension_namespace}.internal.{module}'
    name = obj.__class__.__name__
    if isinstance(obj, FunctionType):
        name = obj.__name__
    torch._C._log_api_usage_once(f'{module}.{name}')



import torch
from torch.autograd.gradcheck import gradcheck

import tvdcn
from utils.deform_conv_test_args import DeformConvTestArgs


def test_deform_conv(dim=2,
                     dtype=torch.float64,
                     device='cuda'):
    torch.manual_seed(12)
    conv_func = getattr(tvdcn, f'deform_conv{dim}d')
    args = DeformConvTestArgs(dim=dim, transposed=False, dtype=dtype, device=device)
    print(args)

    c_output = conv_func(args.input,
                         args.weight,
                         args.offset,
                         args.mask,
                         args.bias,
                         args.stride,
                         args.padding,
                         args.dilation,
                         args.groups)
    c_output.sum().backward()
    c_input_grad = args.input.grad.clone()
    c_weight_grad = args.weight.grad.clone()
    c_offset_grad = args.offset.grad.clone()
    c_mask_grad = args.mask.grad.clone()
    c_bias_grad = args.bias.grad.clone()
    args.zero_grad()
    print(c_output)

    grad_ok = gradcheck(
        lambda inp, wei, off, msk, bi: conv_func(inp, wei, off, msk, bi,
                                                 args.stride,
                                                 args.padding,
                                                 args.dilation,
                                                 args.groups),
        (args.input, args.weight, args.offset, args.mask, args.bias), nondet_tol=args.tol)
    args.zero_grad()
    print('grad_check:', grad_ok)


def test_deform_conv_transpose(dim=2,
                               dtype=torch.float64,
                               device='cuda'):
    torch.manual_seed(12)
    conv_func = getattr(tvdcn, f'deform_conv_transpose{dim}d')
    args = DeformConvTestArgs(dim=dim, transposed=True, dtype=dtype, device=device)
    print(args)

    c_output = conv_func(args.input,
                         args.weight,
                         args.offset,
                         args.mask,
                         args.bias,
                         args.stride,
                         args.padding,
                         args.output_padding,
                         args.dilation,
                         args.groups)
    c_output.sum().backward()
    c_input_grad = args.input.grad.clone()
    c_weight_grad = args.weight.grad.clone()
    c_offset_grad = args.offset.grad.clone()
    c_mask_grad = args.mask.grad.clone()
    c_bias_grad = args.bias.grad.clone()
    args.zero_grad()
    print(c_output)

    grad_ok = gradcheck(
        lambda inp, wei, off, msk, bi: conv_func(inp, wei, off, msk, bi,
                                                 args.stride,
                                                 args.padding,
                                                 args.output_padding,
                                                 args.dilation,
                                                 args.groups),
        (args.input, args.weight, args.offset, args.mask, args.bias), nondet_tol=args.tol)
    args.zero_grad()
    print('grad_check:', grad_ok)


if __name__ == '__main__':
    test_deform_conv()
    test_deform_conv_transpose()



import glob
import os
import sys

import torch
from setuptools import setup
from torch.utils.cpp_extension import CUDAExtension
from torch.utils.cpp_extension import CUDA_HOME
from torch.utils.cpp_extension import CppExtension

PACKAGE_ROOT = 'tvdcn'


def get_version(version_file='_version.py'):
    import importlib.util
    version_file_path = os.path.abspath(os.path.join(PACKAGE_ROOT, version_file))
    try:
        spec = importlib.util.spec_from_file_location('_version', version_file_path)
        version_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(version_module)
        return str(version_module.__version__)
    except:
        return '0.0.0'


def get_parallel_options(backend=None):
    parallel_extra_compile_args = []
    parallel_define_macros = []

    if backend is not None:
        if backend.upper() not in ['OPENMP', 'NATIVE', 'NATIVE_TBB']:
            raise ValueError('Parallel backend options are OPENMP, NATIVE, or NATIVE_TBB. '
                             f'Got unknown backend {backend}.')
    else:  # detect torch parallel backend
        parallel_info_string = torch.__config__.parallel_info()
        parallel_info_array = parallel_info_string.splitlines()
        backend_lines = [line for line in parallel_info_array
                         if line.startswith('ATen parallel backend:')]
        if len(backend_lines):
            backend = backend_lines[0].rsplit(': ')[1]

    backend = backend.lower() if backend is not None else ''
    if backend == 'openmp':
        parallel_define_macros += [('AT_PARALLEL_OPENMP', None)]
        if sys.platform == 'darwin':
            parallel_extra_compile_args.append('-Xpreprocessor')
        parallel_extra_compile_args.append('/openmp' if sys.platform == 'win32' else '-fopenmp')
        if sys.platform == 'darwin':
            parallel_extra_compile_args.append('-lomp')
    elif backend.startswith('native'):
        if backend.endswith('tbb'):
            parallel_define_macros += [('AT_PARALLEL_NATIVE_TBB', None)]
        else:
            parallel_define_macros += [('AT_PARALLEL_NATIVE', None)]
    return parallel_extra_compile_args, parallel_define_macros


def get_extensions():
    extensions_dir = os.path.join(PACKAGE_ROOT, 'csrc')

    main_file = (glob.glob(os.path.join(extensions_dir, '*.cpp')) +
                 glob.glob(os.path.join(extensions_dir, 'ops', '*.cpp'))
                 )

    source_cpu = (glob.glob(os.path.join(extensions_dir, 'ops', 'cpu', '*.cpp')) +
                  glob.glob(os.path.join(extensions_dir, 'ops', 'autograd', '*.cpp')) +
                  glob.glob(os.path.join(extensions_dir, 'ops', 'quantized', 'cpu', '*.cpp')) +
                  glob.glob(os.path.join(extensions_dir, 'ops', 'utils', '*.cpp'))
                  )

    source_cuda = (glob.glob(os.path.join(extensions_dir, 'ops', 'cuda', '*.cu')) +
                   glob.glob(os.path.join(extensions_dir, 'ops', 'cuda', '*.cpp')) +
                   glob.glob(os.path.join(extensions_dir, 'ops', 'autocast', '*.cpp'))
                   )

    sources = main_file + source_cpu
    extension = CppExtension
    extra_compile_args = {'cxx': []}
    extra_compile_args['cxx'].append('/std:c++17' if sys.platform == 'win32' else '-std=c++17')
    define_macros = []

    print('Compiling extensions with following flags:')
    force_cuda = os.getenv('FORCE_CUDA', '0') == '1'
    print(f'  FORCE_CUDA: {force_cuda}')
    debug_mode = os.getenv('DEBUG', '0') == '1'
    print(f'  DEBUG: {debug_mode}')

    nvcc_flags = os.getenv('NVCC_FLAGS', '')
    print(f'  NVCC_FLAGS: {nvcc_flags}')

    # enable cpu parallel
    parallel_extra_compile_args, parallel_define_macros = get_parallel_options('openmp')
    extra_compile_args['cxx'] += parallel_extra_compile_args
    define_macros += parallel_define_macros

    # enable cuda
    if (torch.cuda.is_available() and CUDA_HOME is not None) or force_cuda:
        extension = CUDAExtension
        sources += source_cuda
        define_macros += [('WITH_CUDA', None)]
        if nvcc_flags == '':
            nvcc_flags = []
        else:
            nvcc_flags = nvcc_flags.split(' ')
        nvcc_flags.append('-std=c++17')
        extra_compile_args['nvcc'] = nvcc_flags

    if sys.platform == 'win32':
        define_macros += [(f'{PACKAGE_ROOT}_EXPORTS', None)]
        define_macros += [('USE_PYTHON', None)]
        extra_compile_args['cxx'].append('/MP')

    if debug_mode:
        print('Compiling in debug mode')
        extra_compile_args['cxx'].append('-g')
        extra_compile_args['cxx'].append('-O0')
        if 'nvcc' in extra_compile_args:
            # we have to remove '-OX' and '-g' flag if exists and append
            nvcc_flags = extra_compile_args['nvcc']
            extra_compile_args['nvcc'] = [f for f in nvcc_flags if not ('-O' in f or '-g' in f)]
            extra_compile_args['nvcc'].append('-O0')
            extra_compile_args['nvcc'].append('-g')

    include_dirs = [extensions_dir]
    ext_modules = [
        extension(
            f'{PACKAGE_ROOT}._C',
            sources=sources,
            include_dirs=include_dirs,
            define_macros=define_macros,
            extra_compile_args=extra_compile_args,
        )
    ]
    return ext_modules


def setup_package():
    setup(
        version=get_version(),
        ext_modules=get_extensions(),
        cmdclass={
            'build_ext': torch.utils.cpp_extension.BuildExtension
        },
    )


if __name__ == '__main__':
    setup_package()



from typing import Callable, List, Tuple, Optional, Union

import torch
from torch import nn, Tensor
from torch.nn import init
from torch.nn.common_types import _size_any_t, _size_1_t, _size_2_t, _size_3_t
from torch.nn.modules.utils import _single, _pair, _triple

from .activations import (
    MaskSigmoid, MaskSoftmax1d, MaskSoftmax2d, MaskSoftmax3d,
)
from .deform_conv import _DeformConvNd
from ..extension import _assert_has_ops
from ..utils import _log_api_usage_once

__all__ = [
    'deform_conv_transpose1d',
    'deform_conv_transpose2d',
    'deform_conv_transpose3d',
    'DeformConvTranspose1d',
    'DeformConvTranspose2d',
    'DeformConvTranspose3d',
    'PackedDeformConvTranspose1d',
    'PackedDeformConvTranspose2d',
    'PackedDeformConvTranspose3d',
]


def deform_conv_transpose1d(
        input: Tensor,
        weight: Tensor,
        offset: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        bias: Optional[Tensor] = None,
        stride: _size_1_t = 1,
        padding: _size_1_t = 0,
        output_padding: _size_1_t = 0,
        dilation: _size_1_t = 1,
        groups: int = 1) -> Tensor:
    r"""
    Performs 1D Transposed Deformable Convolution.

    Arguments:
        input (Tensor[batch_size, in_channels, in_width]): input tensor
        weight (Tensor[in_channels, out_channels // groups, kernel_width]):
            convolution weights, split into groups of size (in_channels // groups)
        offset (Tensor[batch_size, offset_groups * kernel_width, in_width]): offsets
            to be applied for each position in the convolution kernel. Default: None
        mask (Tensor[batch_size, mask_groups * kernel_width, in_width]): modulation
            masks to be multiplied with each output of convolution kernel. Default: None
        bias (Tensor[out_channels]): optional bias of shape (out_channels,). Default: None
        stride (int or Tuple[int]): distance between convolution centers. Default: 1
        padding (int or Tuple[int]): padding of zeroes around the input. Default: 0
        output_padding (int or Tuple[int]): additional size added to one side
            of each dimension in the output shape. Default: 0
        dilation (int or Tuple[int]): the spacing between kernel elements. Default: 1
        groups (int): number of blocked connections from input channels to output channels.
            Default: 1

    Returns:
        output (Tensor[batch_sz, out_channels, out_w]): result of convolution

    Examples:
        >>> input = torch.rand(4, 3, 8)
        >>> kw = 3
        >>> weight = torch.rand(3, 5, kw)
        >>> # offset and mask should have the same spatial size as the input.
        >>> offset = torch.rand(4, kw, 8)
        >>> mask = torch.rand(4, kw, 8).sigmoid()
        >>> out = deform_conv_transpose1d(input, weight, offset, mask)
        >>> print(out.shape)
        Output:
        >>>  torch.Size([4, 5, 10])
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(deform_conv_transpose1d)
    _assert_has_ops()

    return torch.ops.tvdcn.deform_conv_transpose1d(
        input, weight, offset, mask, bias,
        stride, padding, output_padding, dilation, groups)


def deform_conv_transpose2d(
        input: Tensor,
        weight: Tensor,
        offset: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        bias: Optional[Tensor] = None,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        output_padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1) -> Tensor:
    r"""
    Performs 2D Transposed Deformable Convolution.

    Arguments:
        input (Tensor[batch_size, in_channels, in_height, in_width]): input tensor
        weight (Tensor[in_channels, out_channels // groups, kernel_height, kernel_width]):
            convolution weights, split into groups of size (in_channels // groups)
        offset (Tensor[batch_size, 2 * offset_groups * kernel_height * kernel_width,
            in_height, in_width]): offsets to be applied for each position in the
            convolution kernel. Default: None
        mask (Tensor[batch_size, mask_groups * kernel_height * kernel_width,
            in_height, in_width]): modulation masks to be multiplied with each output
            of convolution kernel. Default: None
        bias (Tensor[out_channels]): optional bias of shape (out_channels,). Default: None
        stride (int or Tuple[int, int]): distance between convolution centers. Default: 1
        padding (int or Tuple[int, int]): padding of zeroes around the input. Default: 0
        output_padding (int or Tuple[int, int]): additional size added to one side
            of each dimension in the output shape. Default: 0
        dilation (int or Tuple[int, int]): the spacing between kernel elements. Default: 1
        groups (int): number of blocked connections from input channels to output channels.
            Default: 1

    Returns:
        output (Tensor[batch_sz, out_channels, out_h, out_w]): result of convolution

    Examples:
        >>> input = torch.rand(4, 3, 8, 8)
        >>> kh, kw = 3, 3
        >>> weight = torch.rand(3, 5, kh, kw)
        >>> # offset and mask should have the same spatial size as the input.
        >>> offset = torch.rand(4, 2 * kh * kw, 8, 8)
        >>> mask = torch.rand(4, kh * kw, 8, 8).sigmoid()
        >>> out = deform_conv_transpose2d(input, weight, offset, mask)
        >>> print(out.shape)
        Output:
        >>>  torch.Size([4, 5, 10, 10])
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(deform_conv_transpose2d)
    _assert_has_ops()

    return torch.ops.tvdcn.deform_conv_transpose2d(
        input, weight, offset, mask, bias,
        stride, padding, output_padding, dilation, groups)


def deform_conv_transpose3d(
        input: Tensor,
        weight: Tensor,
        offset: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        bias: Optional[Tensor] = None,
        stride: _size_3_t = 1,
        padding: _size_3_t = 0,
        output_padding: _size_3_t = 0,
        dilation: _size_3_t = 1,
        groups: int = 1) -> Tensor:
    r"""
    Performs 3D Transposed Deformable Convolution.

    Arguments:
        input (Tensor[batch_size, in_channels, in_height, in_width, in_depth]): input tensor
        weight (Tensor[in_channels, out_channels // groups, kernel_depth, kernel_height, kernel_width]):
            convolution weights, split into groups of size (in_channels // groups)
        offset (Tensor[batch_size, 3 * offset_groups * kernel_depth * kernel_height * kernel_width,
            in_depth, in_height, in_width]): offsets to be applied for each position in the
            convolution kernel. Default: None
        mask (Tensor[batch_size, mask_groups * kernel_depth * kernel_height * kernel_width,
            in_depth, in_height, in_width]): modulation masks to be multiplied with each output
            of convolution kernel. Default: None
        bias (Tensor[out_channels]): optional bias of shape (out_channels,). Default: None
        stride (int or Tuple[int, int, int]): distance between convolution centers. Default: 1
        padding (int or Tuple[int, int, int]): height/width of padding of zeroes around
            each image. Default: 0
        output_padding (int or Tuple[int, int, int]): additional size added to one side
            of each dimension in the output shape. Default: 0
        dilation (int or Tuple[int, int, int]): the spacing between kernel elements. Default: 1
        groups (int): number of blocked connections from input channels to output channels.
            Default: 1

    Returns:
        output (Tensor[batch_sz, out_channels, out_d, out_h, out_w]): result of convolution

    Examples:
        >>> input = torch.rand(4, 3, 8, 8, 8)
        >>> kd, kh, kw = 3, 3, 3
        >>> weight = torch.rand(3, 5, kd, kh, kw)
        >>> # offset and mask should have the same spatial size as the input.
        >>> offset = torch.rand(4, 3 * kd * kh * kw, 8, 8, 8)
        >>> mask = torch.rand(4, kd * kh * kw, 8, 8, 8).sigmoid()
        >>> out = deform_conv_transpose3d(input, weight, offset, mask)
        >>> print(out.shape)
        Output:
        >>> torch.Size([4, 5, 10, 10, 10])
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(deform_conv_transpose3d)
    _assert_has_ops()

    return torch.ops.tvdcn.deform_conv_transpose3d(
        input, weight, offset, mask, bias,
        stride, padding, output_padding, dilation, groups)


################################################################################
# Modules
################################################################################
# noinspection PyMethodOverriding
class _DeformConvTransposeNd(_DeformConvNd):
    """
    Base class for DeformConvTranspose
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_any_t,
                 stride: _size_any_t,
                 padding: Union[str, _size_any_t],
                 dilation: _size_any_t,
                 transposed: bool,
                 output_padding: _size_any_t,
                 groups: int,
                 bias: bool,
                 padding_mode: str,
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__(
            in_channels, out_channels, kernel_size, stride,
            padding, dilation, transposed, output_padding,
            groups, bias, padding_mode, **factory_kwargs)

    def _output_padding(self, input: Tensor, output_size: Optional[List[int]],
                        stride: List[int], padding: List[int], kernel_size: List[int],
                        num_spatial_dims: int, dilation: Optional[List[int]] = None) -> List[int]:
        if output_size is None:
            ret = _single(self.output_padding)  # converting to list if was not already
        else:
            has_batch_dim = input.dim() == num_spatial_dims + 2
            num_non_spatial_dims = 2 if has_batch_dim else 1
            if len(output_size) == num_non_spatial_dims + num_spatial_dims:
                output_size = output_size[num_non_spatial_dims:]
            if len(output_size) != num_spatial_dims:
                raise ValueError(
                    "ConvTranspose{}D: for {}D input, output_size must have {} or {} elements (got {})"
                    .format(num_spatial_dims, input.dim(), num_spatial_dims,
                            num_non_spatial_dims + num_spatial_dims, len(output_size)))

            min_sizes = torch.jit.annotate(List[int], [])
            max_sizes = torch.jit.annotate(List[int], [])
            for d in range(num_spatial_dims):
                dim_size = ((input.size(d + num_non_spatial_dims) - 1) * stride[d] -
                            2 * padding[d] +
                            (dilation[d] if dilation is not None else 1) * (kernel_size[d] - 1) + 1)
                min_sizes.append(dim_size)
                max_sizes.append(min_sizes[d] + stride[d] - 1)

            for i in range(len(output_size)):
                size = output_size[i]
                min_size = min_sizes[i]
                max_size = max_sizes[i]
                if size < min_size or size > max_size:
                    raise ValueError((
                        "requested an output size of {}, but valid sizes range "
                        "from {} to {} (for an input of {})").format(
                        output_size, min_sizes, max_sizes, input.size()[2:]))

            res = torch.jit.annotate(List[int], [])
            for d in range(num_spatial_dims):
                res.append(output_size[d] - min_sizes[d])

            ret = res
        return ret

    def _conv_transpose_forward(self,
                                input: Tensor,
                                weight: Tensor,
                                offset: Optional[Tensor],
                                mask: Optional[Tensor],
                                bias: Optional[Tensor],
                                output_size: Optional[List[int]] = None) -> Tensor:
        raise NotImplementedError

    def forward(self,
                input: Tensor,
                offset: Optional[Tensor],
                mask: Optional[Tensor] = None,
                output_size: Optional[List[int]] = None) -> Tensor:
        return self._conv_transpose_forward(input, self.weight, offset, mask, self.bias, output_size)


class DeformConvTranspose1d(_DeformConvTransposeNd):
    """
    See :func:`deform_conv_transpose1d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_1_t,
                 stride: _size_1_t = 1,
                 padding: Union[str, _size_1_t] = 0,
                 dilation: _size_1_t = 1,
                 output_padding: _size_1_t = 0,
                 groups: int = 1,
                 bias: bool = True,
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = _single(padding)
        dilation = _single(dilation)
        output_padding = _single(output_padding)
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            True, output_padding, groups, bias, padding_mode, **factory_kwargs)

    def _conv_transpose_forward(self,
                                input: Tensor,
                                weight: Tensor,
                                offset: Optional[Tensor],
                                mask: Optional[Tensor],
                                bias: Optional[Tensor],
                                output_size: Optional[List[int]] = None) -> Tensor:
        if self.padding_mode != 'zeros':
            raise ValueError('Only `zeros` padding mode is supported for DeformConvTranspose1d.')

        assert isinstance(self.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        num_spatial_dims = 1
        output_padding = self._output_padding(
            input, output_size, self.stride, self.padding, self.kernel_size,  # type: ignore[arg-type]
            num_spatial_dims, self.dilation)  # type: ignore[arg-type]
        return deform_conv_transpose1d(input, weight, offset, mask, bias,
                                       self.stride, self.padding,
                                       (output_padding[0],),
                                       self.dilation, self.groups)


class DeformConvTranspose2d(_DeformConvTransposeNd):
    """
    See :func:`deform_conv_transpose2d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_2_t,
                 stride: _size_2_t = 1,
                 padding: Union[str, _size_2_t] = 0,
                 dilation: _size_2_t = 1,
                 output_padding: _size_2_t = 0,
                 groups: int = 1,
                 bias: bool = True,
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        output_padding = _pair(output_padding)
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            True, output_padding, groups, bias, padding_mode, **factory_kwargs)

    def _conv_transpose_forward(self,
                                input: Tensor,
                                weight: Tensor,
                                offset: Optional[Tensor],
                                mask: Optional[Tensor],
                                bias: Optional[Tensor],
                                output_size: Optional[List[int]] = None) -> Tensor:
        if self.padding_mode != 'zeros':
            raise ValueError('Only `zeros` padding mode is supported for DeformConvTranspose2d.')

        assert isinstance(self.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        num_spatial_dims = 2
        output_padding = self._output_padding(
            input, output_size, self.stride, self.padding, self.kernel_size,  # type: ignore[arg-type]
            num_spatial_dims, self.dilation)  # type: ignore[arg-type]

        return deform_conv_transpose2d(input, weight, offset, mask, bias,
                                       self.stride, self.padding,  # type: ignore[arg-type]
                                       (output_padding[0], output_padding[1]),
                                       self.dilation, self.groups)  # type: ignore[arg-type]


class DeformConvTranspose3d(_DeformConvTransposeNd):
    """
    See :func:`deform_conv_transpose3d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_3_t,
                 stride: _size_3_t = 1,
                 padding: Union[str, _size_3_t] = 0,
                 dilation: _size_3_t = 1,
                 output_padding: _size_3_t = 0,
                 groups: int = 1,
                 bias: bool = True,
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = _triple(padding)
        dilation = _triple(dilation)
        output_padding = _triple(output_padding)
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            True, output_padding, groups, bias, padding_mode, **factory_kwargs)

    def _conv_transpose_forward(self,
                                input: Tensor,
                                weight: Tensor,
                                offset: Optional[Tensor],
                                mask: Optional[Tensor],
                                bias: Optional[Tensor],
                                output_size: Optional[List[int]] = None) -> Tensor:
        if self.padding_mode != 'zeros':
            raise ValueError('Only `zeros` padding mode is supported for DeformConvTranspose3d.')

        assert isinstance(self.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        num_spatial_dims = 3
        output_padding = self._output_padding(
            input, output_size, self.stride, self.padding, self.kernel_size,  # type: ignore[arg-type]
            num_spatial_dims, self.dilation)  # type: ignore[arg-type]

        return deform_conv_transpose3d(input, weight, offset, mask, bias,
                                       self.stride, self.padding,  # type: ignore[arg-type]
                                       (output_padding[0], output_padding[1], output_padding[2]),
                                       self.dilation, self.groups)  # type: ignore[arg-type]


################################################################################
# Packed Modules
################################################################################
# noinspection PyMethodOverriding
class PackedDeformConvTranspose1d(DeformConvTranspose1d):
    """
    Packed version of :class:`DeformConvTranspose1d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_1_t,
                 stride: _size_1_t = 1,
                 padding: Union[str, _size_1_t] = 0,
                 dilation: _size_1_t = 1,
                 output_padding: _size_1_t = 0,
                 groups: int = 1,
                 offset_groups: int = 1,
                 mask_groups: int = 1,
                 bias: bool = True,
                 generator_bias: bool = False,
                 deformable: bool = True,
                 modulated: bool = False,
                 offset_activation: Optional[Union[str, Callable]] = None,
                 mask_activation: Optional[Union[str, Callable]] = 'sigmoid',
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            output_padding, groups, bias, padding_mode, device, dtype)

        if in_channels % offset_groups != 0:
            raise ValueError('in_channels must be divisible by offset_groups')
        if out_channels % offset_groups != 0:
            raise ValueError('out_channels must be divisible by offset_groups')

        if in_channels % mask_groups != 0:
            raise ValueError('in_channels must be divisible by mask_groups')
        if out_channels % mask_groups != 0:
            raise ValueError('out_channels must be divisible by mask_groups')

        if isinstance(offset_activation, str):
            raise ValueError('currently, no activation is supported for offset')
        if isinstance(mask_activation, str):
            if mask_activation == 'sigmoid':
                mask_activation = MaskSigmoid(scale=2.)
            elif mask_activation == 'softmax':
                mask_activation = MaskSoftmax1d(self.kernel_size)  # type: ignore[arg-type]
            else:
                raise ValueError('only \"sigmoid\" and \"softmax\" activations are supported for mask')

        self.offset_groups = offset_groups
        self.mask_groups = mask_groups

        self.deformable = deformable
        self.modulated = modulated

        if self.deformable:
            self.offset_generator = nn.Conv1d(
                self.in_channels,
                self.kernel_size[0] * self.offset_groups,
                kernel_size=1,
                groups=self.offset_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('offset_generator', None)

        if self.modulated:
            self.mask_generator = nn.Conv1d(
                self.in_channels,
                self.kernel_size[0] * self.mask_groups,
                kernel_size=1,
                groups=self.mask_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('mask_generator', None)

        self.offset_activation = offset_activation
        self.mask_activation = mask_activation

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not hasattr(self, 'modulated'):
            return
        super().reset_parameters()
        if self.offset_generator is not None:
            init.zeros_(self.offset_generator.weight)
            if self.offset_generator.bias is not None:
                init.zeros_(self.offset_generator.bias)
        if self.mask_generator is not None:
            init.zeros_(self.mask_generator.weight)
            if self.mask_generator.bias is not None:
                init.zeros_(self.mask_generator.bias)

    def forward(self, input: Tensor, output_size: Optional[List[int]] = None) -> Tensor:
        if self.deformable and self.offset_generator is not None:
            offset = self.offset_generator(input)
            if self.offset_activation is not None:
                offset = self.offset_activation(offset)
        else:
            offset = None

        if self.modulated and self.mask_generator is not None:
            mask = self.mask_generator(input)
            if self.mask_activation is not None:
                mask = self.mask_activation(mask)
        else:
            mask = None

        return self._conv_transpose_forward(input, self.weight, offset, mask, self.bias, output_size)


# noinspection PyMethodOverriding
class PackedDeformConvTranspose2d(DeformConvTranspose2d):
    """
    Packed version of :class:`DeformConvTranspose2d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_2_t,
                 stride: _size_2_t = 1,
                 padding: Union[str, _size_2_t] = 0,
                 dilation: _size_2_t = 1,
                 output_padding: _size_2_t = 0,
                 groups: int = 1,
                 offset_groups: int = 1,
                 mask_groups: int = 1,
                 bias: bool = True,
                 generator_bias: bool = False,
                 deformable: bool = True,
                 modulated: bool = False,
                 offset_activation: Optional[Union[str, Callable]] = None,
                 mask_activation: Optional[Union[str, Callable]] = 'sigmoid',
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            output_padding, groups, bias, padding_mode, device, dtype)

        if in_channels % offset_groups != 0:
            raise ValueError('in_channels must be divisible by offset_groups')
        if out_channels % offset_groups != 0:
            raise ValueError('out_channels must be divisible by offset_groups')

        if in_channels % mask_groups != 0:
            raise ValueError('in_channels must be divisible by mask_groups')
        if out_channels % mask_groups != 0:
            raise ValueError('out_channels must be divisible by mask_groups')

        if isinstance(offset_activation, str):
            raise ValueError('currently, no activation is supported for offset')
        if isinstance(mask_activation, str):
            if mask_activation == 'sigmoid':
                mask_activation = MaskSigmoid(scale=2.)
            elif mask_activation == 'softmax':
                mask_activation = MaskSoftmax2d(self.kernel_size)  # type: ignore[arg-type]
            else:
                raise ValueError('only \"sigmoid\" and \"softmax\" activations are supported for mask')

        self.offset_groups = offset_groups
        self.mask_groups = mask_groups

        self.deformable = deformable
        self.modulated = modulated

        if self.deformable:
            self.offset_generator = nn.Conv2d(
                self.in_channels,
                2 * self.kernel_size[0] * self.kernel_size[1] * self.offset_groups,
                kernel_size=1,
                groups=self.offset_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('offset_generator', None)

        if self.modulated:
            self.mask_generator = nn.Conv2d(
                self.in_channels,
                self.kernel_size[0] * self.kernel_size[1] * self.mask_groups,
                kernel_size=1,
                groups=self.mask_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('mask_generator', None)

        self.offset_activation = offset_activation
        self.mask_activation = mask_activation

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not hasattr(self, 'modulated'):
            return
        super().reset_parameters()
        if self.offset_generator is not None:
            init.zeros_(self.offset_generator.weight)
            if self.offset_generator.bias is not None:
                init.zeros_(self.offset_generator.bias)
        if self.mask_generator is not None:
            init.zeros_(self.mask_generator.weight)
            if self.mask_generator.bias is not None:
                init.zeros_(self.mask_generator.bias)

    def forward(self, input: Tensor, output_size: Optional[List[int]] = None) -> Tensor:
        if self.deformable and self.offset_generator is not None:
            offset = self.offset_generator(input)
            if self.offset_activation is not None:
                offset = self.offset_activation(offset)
        else:
            offset = None

        if self.modulated and self.mask_generator is not None:
            mask = self.mask_generator(input)
            if self.mask_activation is not None:
                mask = self.mask_activation(mask)
        else:
            mask = None

        return self._conv_transpose_forward(input, self.weight, offset, mask, self.bias, output_size)


# noinspection PyMethodOverriding
class PackedDeformConvTranspose3d(DeformConvTranspose3d):
    """
    Packed version of :class:`DeformConvTranspose3d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_3_t,
                 stride: _size_3_t = 1,
                 padding: Union[str, _size_3_t] = 0,
                 dilation: _size_3_t = 1,
                 output_padding: _size_3_t = 0,
                 groups: int = 1,
                 offset_groups: int = 1,
                 mask_groups: int = 1,
                 bias: bool = True,
                 generator_bias: bool = False,
                 deformable: bool = True,
                 modulated: bool = False,
                 offset_activation: Optional[Union[str, Callable]] = None,
                 mask_activation: Optional[Union[str, Callable]] = 'sigmoid',
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            output_padding, groups, bias, padding_mode, device, dtype)

        if in_channels % offset_groups != 0:
            raise ValueError('in_channels must be divisible by offset_groups')
        if out_channels % offset_groups != 0:
            raise ValueError('out_channels must be divisible by offset_groups')

        if in_channels % mask_groups != 0:
            raise ValueError('in_channels must be divisible by mask_groups')
        if out_channels % mask_groups != 0:
            raise ValueError('out_channels must be divisible by mask_groups')

        if isinstance(offset_activation, str):
            raise ValueError('currently, no activation is supported for offset')
        if isinstance(mask_activation, str):
            if mask_activation == 'sigmoid':
                mask_activation = MaskSigmoid(scale=2.)
            elif mask_activation == 'softmax':
                mask_activation = MaskSoftmax3d(self.kernel_size)  # type: ignore[arg-type]
            else:
                raise ValueError('only \"sigmoid\" and \"softmax\" activations are supported for mask')

        self.offset_groups = offset_groups
        self.mask_groups = mask_groups

        self.deformable = deformable
        self.modulated = modulated

        if self.deformable:
            self.offset_generator = nn.Conv3d(
                self.in_channels,
                3 * self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2] * self.offset_groups,
                kernel_size=1,
                groups=self.offset_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('offset_generator', None)

        if self.modulated:
            self.mask_generator = nn.Conv3d(
                self.in_channels,
                self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2] * self.mask_groups,
                kernel_size=1,
                groups=self.mask_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('mask_generator', None)

        self.offset_activation = offset_activation
        self.mask_activation = mask_activation

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not hasattr(self, 'modulated'):
            return
        super().reset_parameters()
        if self.offset_generator is not None:
            init.zeros_(self.offset_generator.weight)
            if self.offset_generator.bias is not None:
                init.zeros_(self.offset_generator.bias)
        if self.mask_generator is not None:
            init.zeros_(self.mask_generator.weight)
            if self.mask_generator.bias is not None:
                init.zeros_(self.mask_generator.bias)

    def forward(self, input: Tensor, output_size: Optional[List[int]] = None) -> Tensor:
        if self.deformable and self.offset_generator is not None:
            offset = self.offset_generator(input)
            if self.offset_activation is not None:
                offset = self.offset_activation(offset)
        else:
            offset = None

        if self.modulated and self.mask_generator is not None:
            mask = self.mask_generator(input)
            if self.mask_activation is not None:
                mask = self.mask_activation(mask)
        else:
            mask = None

        return self._conv_transpose_forward(input, self.weight, offset, mask, self.bias, output_size)



from typing import Callable, Union, Optional, Tuple

import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch.nn import init
from torch.nn.common_types import _size_any_t, _size_1_t, _size_2_t, _size_3_t
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _single, _pair, _triple

from .activations import (
    MaskSigmoid, MaskSoftmax1d, MaskSoftmax2d, MaskSoftmax3d,
)
from ..extension import _assert_has_ops
from ..utils import _log_api_usage_once

__all__ = [
    'deform_conv1d',
    'deform_conv2d',
    'deform_conv3d',
    'DeformConv1d',
    'DeformConv2d',
    'DeformConv3d',
    'PackedDeformConv1d',
    'PackedDeformConv2d',
    'PackedDeformConv3d',
]


def deform_conv1d(
        input: Tensor,
        weight: Tensor,
        offset: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        bias: Optional[Tensor] = None,
        stride: _size_1_t = 1,
        padding: _size_1_t = 0,
        dilation: _size_1_t = 1,
        groups: int = 1) -> Tensor:
    r"""
    Performs 1D Deformable Convolution.

    Arguments:
        input (Tensor[batch_size, in_channels, in_width]): input tensor
        weight (Tensor[out_channels, in_channels // groups, kernel_width]):
            convolution weights, split into groups of size (in_channels // groups)
        offset (Tensor[batch_size, offset_groups * kernel_width, out_width]): offsets
            to be applied for each position in the convolution kernel. Default: None
        mask (Tensor[batch_size, mask_groups * kernel_width, out_width]): modulation
            masks to be multiplied with each output of convolution kernel. Default: None
        bias (Tensor[out_channels]): optional bias of shape (out_channels,). Default: None
        stride (int or Tuple[int]): distance between convolution centers. Default: 1
        padding (int or Tuple[int]): padding of zeroes around the input. Default: 0
        dilation (int or Tuple[int]): the spacing between kernel elements. Default: 1
        groups (int): number of blocked connections from input channels to output channels.
            Default: 1

    Returns:
        output (Tensor[batch_sz, out_channels, out_w]): result of convolution

    Examples:
        >>> input = torch.rand(4, 3, 10)
        >>> kw = 3
        >>> weight = torch.rand(5, 3, kw)
        >>> # offset and mask should have the same spatial size as the output.
        >>> # In this case, for an input of 10, stride of 1 and kernel size of 3,
        >>> # without padding, the output size is 8.
        >>> offset = torch.rand(4, kw, 8)
        >>> mask = torch.rand(4, kw, 8).sigmoid()
        >>> out = deform_conv1d(input, weight, offset, mask)
        >>> print(out.shape)
        Output:
        >>>  torch.Size([4, 5, 8])
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(deform_conv1d)
    _assert_has_ops()

    return torch.ops.tvdcn.deform_conv1d(
        input, weight, offset, mask, bias,
        stride, padding, dilation, groups)


def deform_conv2d(
        input: Tensor,
        weight: Tensor,
        offset: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        bias: Optional[Tensor] = None,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1) -> Tensor:
    r"""
    Performs 2D Deformable Convolution.

    References:
        https://arxiv.org/abs/1811.11168
        https://arxiv.org/abs/1703.06211
        https://arxiv.org/abs/2211.05778

    Arguments:
        input (Tensor[batch_size, in_channels, in_height, in_width]): input tensor
        weight (Tensor[out_channels, in_channels // groups, kernel_height, kernel_width]):
            convolution weights, split into groups of size (in_channels // groups)
        offset (Tensor[batch_size, 2 * offset_groups * kernel_height * kernel_width,
            out_height, out_width]): offsets to be applied for each position in the
            convolution kernel. Default: None
        mask (Tensor[batch_size, mask_groups * kernel_height * kernel_width,
            out_height, out_width]): modulation masks to be multiplied with each output
            of convolution kernel. Default: None
        bias (Tensor[out_channels]): optional bias of shape (out_channels,). Default: None
        stride (int or Tuple[int, int]): distance between convolution centers. Default: 1
        padding (int or Tuple[int, int]): padding of zeroes around the input. Default: 0
        dilation (int or Tuple[int, int]): the spacing between kernel elements. Default: 1
        groups (int): number of blocked connections from input channels to output channels.
            Default: 1

    Returns:
        output (Tensor[batch_sz, out_channels, out_h, out_w]): result of convolution

    Examples:
        >>> input = torch.rand(4, 3, 10, 10)
        >>> kh, kw = 3, 3
        >>> weight = torch.rand(5, 3, kh, kw)
        >>> # offset and mask should have the same spatial size as the output.
        >>> # In this case, for an input of 10, stride of 1 and kernel size of 3,
        >>> # without padding, the output size is 8.
        >>> offset = torch.rand(4, 2 * kh * kw, 8, 8)
        >>> mask = torch.rand(4, kh * kw, 8, 8).sigmoid()
        >>> out = deform_conv2d(input, weight, offset, mask)
        >>> print(out.shape)
        Output:
        >>>  torch.Size([4, 5, 8, 8])
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(deform_conv2d)
    _assert_has_ops()

    return torch.ops.tvdcn.deform_conv2d(
        input, weight, offset, mask, bias,
        stride, padding, dilation, groups)


def deform_conv3d(
        input: Tensor,
        weight: Tensor,
        offset: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        bias: Optional[Tensor] = None,
        stride: _size_3_t = 1,
        padding: _size_3_t = 0,
        dilation: _size_3_t = 1,
        groups: int = 1) -> Tensor:
    r"""
    Performs 3D Deformable Convolution.

    Arguments:
        input (Tensor[batch_size, in_channels, in_height, in_width, in_depth]): input tensor
        weight (Tensor[out_channels, in_channels // groups, kernel_depth, kernel_height, kernel_width]):
            convolution weights, split into groups of size (in_channels // groups)
        offset (Tensor[batch_size, 3 * offset_groups * kernel_depth * kernel_height * kernel_width,
            out_depth, out_height, out_width]): offsets to be applied for each position in the
            convolution kernel. Default: None
        mask (Tensor[batch_size, mask_groups * kernel_depth * kernel_height * kernel_width,
            out_depth, out_height, out_width]): modulation masks to be multiplied with each output
            of convolution kernel. Default: None
        bias (Tensor[out_channels]): optional bias of shape (out_channels,). Default: None
        stride (int or Tuple[int, int, int]): distance between convolution centers. Default: 1
        padding (int or Tuple[int, int, int]): height/width of padding of zeroes around
            each image. Default: 0
        dilation (int or Tuple[int, int, int]): the spacing between kernel elements. Default: 1
        groups (int): number of blocked connections from input channels to output channels.
            Default: 1

    Returns:
        output (Tensor[batch_sz, out_channels, out_d, out_h, out_w]): result of convolution

    Examples:
        >>> input = torch.rand(4, 3, 10, 10, 10)
        >>> kd, kh, kw = 3, 3, 3
        >>> weight = torch.rand(5, 3, kd, kh, kw)
        >>> # offset and mask should have the same spatial size as the output.
        >>> # In this case, for an input of 10, stride of 1 and kernel size of 3,
        >>> # without padding, the output size is 8.
        >>> offset = torch.rand(4, 3 * kd * kh * kw, 8, 8, 8)
        >>> mask = torch.rand(4, kd * kh * kw, 8, 8, 8).sigmoid()
        >>> out = deform_conv3d(input, weight, offset, mask)
        >>> print(out.shape)
        Output:
        >>> torch.Size([4, 5, 8, 8, 8])
    """
    if not torch.jit.is_scripting() and not torch.jit.is_tracing():
        _log_api_usage_once(deform_conv3d)
    _assert_has_ops()

    return torch.ops.tvdcn.deform_conv3d(
        input, weight, offset, mask, bias,
        stride, padding, dilation, groups)


################################################################################
# Modules
################################################################################
# noinspection PyMethodOverriding
class _DeformConvNd(_ConvNd):
    """
    Base class for DeformConv
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_any_t,
                 stride: _size_any_t,
                 padding: Union[str, _size_any_t],
                 dilation: _size_any_t,
                 transposed: bool,
                 output_padding: _size_any_t,
                 groups: int,
                 bias: bool,
                 padding_mode: str,
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__(
            in_channels, out_channels, kernel_size, stride,
            padding, dilation, transposed, output_padding,
            groups, bias, padding_mode, **factory_kwargs)

    def _conv_forward(self,
                      input: Tensor,
                      weight: Tensor,
                      offset: Optional[Tensor],
                      mask: Optional[Tensor],
                      bias: Optional[Tensor]) -> Tensor:
        raise NotImplementedError

    def forward(self, input: Tensor, offset: Optional[Tensor], mask: Optional[Tensor] = None) -> Tensor:
        return self._conv_forward(input, self.weight, offset, mask, self.bias)

    def extra_repr(self):
        s = ('{in_channels}, {out_channels}, kernel_size={kernel_size}'
             ', stride={stride}')
        if self.padding != (0,) * len(self.padding):
            s += ', padding={padding}'
        if self.dilation != (1,) * len(self.dilation):
            s += ', dilation={dilation}'
        if self.output_padding != (0,) * len(self.output_padding):
            s += ', output_padding={output_padding}'
        if self.groups != 1:
            s += ', groups={groups}'
        if hasattr(self, 'offset_groups') and self.offset_groups != 1:
            s += ', offset_groups={offset_groups}'
        if hasattr(self, 'mask_groups') and self.mask_groups != 1:
            s += ', mask_groups={mask_groups}'
        if self.bias is None:
            s += ', bias=False'
        if hasattr(self, 'deformable') and not self.deformable:
            s += ', deformable=False'
        if hasattr(self, 'modulated') and not self.modulated:
            s += ', modulated=False'
        if self.padding_mode != 'zeros':
            s += ', padding_mode={padding_mode}'
        return s.format(**self.__dict__)


class DeformConv1d(_DeformConvNd):
    """
    See :func:`deform_conv1d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_1_t,
                 stride: _size_1_t = 1,
                 padding: Union[str, _size_1_t] = 0,
                 dilation: _size_1_t = 1,
                 groups: int = 1,
                 bias: bool = True,
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size_ = _single(kernel_size)
        stride_ = _single(stride)
        padding_ = padding if isinstance(padding, str) else _single(padding)
        dilation_ = _single(dilation)
        super().__init__(
            in_channels, out_channels, kernel_size_, stride_, padding_, dilation_,
            False, _single(0), groups, bias, padding_mode, **factory_kwargs)

    def _conv_forward(self,
                      input: Tensor,
                      weight: Tensor,
                      offset: Optional[Tensor],
                      mask: Optional[Tensor],
                      bias: Optional[Tensor]) -> Tensor:
        if self.padding_mode != 'zeros':
            return deform_conv1d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                                 weight, offset, mask, bias, self.stride,
                                 (0,), self.dilation, self.groups)
        return deform_conv1d(input, weight, offset, mask, bias,
                             self.stride, self.padding, self.dilation, self.groups)


class DeformConv2d(_DeformConvNd):
    """
    See :func:`deform_conv2d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_2_t,
                 stride: _size_2_t = 1,
                 padding: Union[str, _size_2_t] = 0,
                 dilation: _size_2_t = 1,
                 groups: int = 1,
                 bias: bool = True,
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size_ = _pair(kernel_size)
        stride_ = _pair(stride)
        padding_ = padding if isinstance(padding, str) else _pair(padding)
        dilation_ = _pair(dilation)
        super().__init__(
            in_channels, out_channels, kernel_size_, stride_, padding_, dilation_,
            False, _pair(0), groups, bias, padding_mode, **factory_kwargs)

    def _conv_forward(self,
                      input: Tensor,
                      weight: Tensor,
                      offset: Optional[Tensor],
                      mask: Optional[Tensor],
                      bias: Optional[Tensor]) -> Tensor:
        if self.padding_mode != 'zeros':
            return deform_conv2d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                                 weight, offset, mask, bias, self.stride,  # type: ignore[arg-type]
                                 (0, 0), self.dilation, self.groups)  # type: ignore[arg-type]
        return deform_conv2d(input, weight, offset, mask, bias,
                             self.stride, self.padding, self.dilation, self.groups)  # type: ignore[arg-type]


class DeformConv3d(_DeformConvNd):
    """
    See :func:`deform_conv3d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_3_t,
                 stride: _size_3_t = 1,
                 padding: Union[str, _size_3_t] = 0,
                 dilation: _size_3_t = 1,
                 groups: int = 1,
                 bias: bool = True,
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        kernel_size_ = _triple(kernel_size)
        stride_ = _triple(stride)
        padding_ = padding if isinstance(padding, str) else _triple(padding)
        dilation_ = _triple(dilation)
        super().__init__(
            in_channels, out_channels, kernel_size_, stride_, padding_, dilation_,
            False, _triple(0), groups, bias, padding_mode, **factory_kwargs)

    def _conv_forward(self,
                      input: Tensor,
                      weight: Tensor,
                      offset: Optional[Tensor],
                      mask: Optional[Tensor],
                      bias: Optional[Tensor]) -> Tensor:
        if self.padding_mode != 'zeros':
            return deform_conv3d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                                 weight, offset, mask, bias, self.stride,  # type: ignore[arg-type]
                                 (0, 0, 0), self.dilation, self.groups)  # type: ignore[arg-type]
        return deform_conv3d(input, weight, offset, mask, bias,
                             self.stride, self.padding, self.dilation, self.groups)  # type: ignore[arg-type]


################################################################################
# Packed Modules
################################################################################
# noinspection PyMethodOverriding
class PackedDeformConv1d(DeformConv1d):
    """
    Packed version of :class:`DeformConv1d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_1_t,
                 stride: _size_1_t = 1,
                 padding: Union[str, _size_1_t] = 0,
                 dilation: _size_1_t = 1,
                 groups: int = 1,
                 offset_groups: int = 1,
                 mask_groups: int = 1,
                 bias: bool = True,
                 generator_bias: bool = False,
                 deformable: bool = True,
                 modulated: bool = False,
                 offset_activation: Optional[Union[str, Callable]] = None,
                 mask_activation: Optional[Union[str, Callable]] = 'sigmoid',
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            groups, bias, padding_mode, device, dtype)

        if in_channels % offset_groups != 0:
            raise ValueError('in_channels must be divisible by offset_groups')
        if out_channels % offset_groups != 0:
            raise ValueError('out_channels must be divisible by offset_groups')

        if in_channels % mask_groups != 0:
            raise ValueError('in_channels must be divisible by mask_groups')
        if out_channels % mask_groups != 0:
            raise ValueError('out_channels must be divisible by mask_groups')

        if isinstance(offset_activation, str):
            raise ValueError('currently, no activation is supported for offset')
        if isinstance(mask_activation, str):
            if mask_activation == 'sigmoid':
                mask_activation = MaskSigmoid(scale=2.)
            elif mask_activation == 'softmax':
                mask_activation = MaskSoftmax1d(self.kernel_size)  # type: ignore[arg-type]
            else:
                raise ValueError('only \"sigmoid\" and \"softmax\" activations are supported for mask')

        self.offset_groups = offset_groups
        self.mask_groups = mask_groups

        self.deformable = deformable
        self.modulated = modulated

        if self.deformable:
            self.offset_generator = nn.Conv1d(
                self.in_channels,
                self.kernel_size[0] * self.offset_groups,
                kernel_size=self.kernel_size,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.offset_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('offset_generator', None)

        if self.modulated:
            self.mask_generator = nn.Conv1d(
                self.in_channels,
                self.kernel_size[0] * self.mask_groups,
                kernel_size=self.kernel_size,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.mask_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('mask_generator', None)

        self.offset_activation = offset_activation
        self.mask_activation = mask_activation

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not hasattr(self, 'modulated'):
            return
        super().reset_parameters()
        if self.offset_generator is not None:
            init.zeros_(self.offset_generator.weight)
            if self.offset_generator.bias is not None:
                init.zeros_(self.offset_generator.bias)
        if self.mask_generator is not None:
            init.zeros_(self.mask_generator.weight)
            if self.mask_generator.bias is not None:
                init.zeros_(self.mask_generator.bias)

    def forward(self, input: Tensor) -> Tensor:
        if self.deformable and self.offset_generator is not None:
            offset = self.offset_generator(input)
            if self.offset_activation is not None:
                offset = self.offset_activation(offset)
        else:
            offset = None

        if self.modulated and self.mask_generator is not None:
            mask = self.mask_generator(input)
            if self.mask_activation is not None:
                mask = self.mask_activation(mask)
        else:
            mask = None

        return self._conv_forward(input, self.weight, offset, mask, self.bias)


# noinspection PyMethodOverriding
class PackedDeformConv2d(DeformConv2d):
    """
    Packed version of :class:`DeformConv2d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_2_t,
                 stride: _size_2_t = 1,
                 padding: Union[str, _size_2_t] = 0,
                 dilation: _size_2_t = 1,
                 groups: int = 1,
                 offset_groups: int = 1,
                 mask_groups: int = 1,
                 bias: bool = True,
                 generator_bias: bool = False,
                 deformable: bool = True,
                 modulated: bool = False,
                 offset_activation: Optional[Union[str, Callable]] = None,
                 mask_activation: Optional[Union[str, Callable]] = 'sigmoid',
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            groups, bias, padding_mode, device, dtype)

        if in_channels % offset_groups != 0:
            raise ValueError('in_channels must be divisible by offset_groups')
        if out_channels % offset_groups != 0:
            raise ValueError('out_channels must be divisible by offset_groups')

        if in_channels % mask_groups != 0:
            raise ValueError('in_channels must be divisible by mask_groups')
        if out_channels % mask_groups != 0:
            raise ValueError('out_channels must be divisible by mask_groups')

        if isinstance(offset_activation, str):
            raise ValueError('currently, no activation is supported for offset')
        if isinstance(mask_activation, str):
            if mask_activation == 'sigmoid':
                mask_activation = MaskSigmoid(scale=2.)
            elif mask_activation == 'softmax':
                mask_activation = MaskSoftmax2d(self.kernel_size)  # type: ignore[arg-type]
            else:
                raise ValueError('only \"sigmoid\" and \"softmax\" activations are supported for mask')

        self.offset_groups = offset_groups
        self.mask_groups = mask_groups

        self.deformable = deformable
        self.modulated = modulated

        if self.deformable:
            self.offset_generator = nn.Conv2d(
                self.in_channels,
                2 * self.kernel_size[0] * self.kernel_size[1] * self.offset_groups,
                kernel_size=self.kernel_size,  # type: ignore[arg-type]
                stride=self.stride,  # type: ignore[arg-type]
                padding=self.padding,
                dilation=self.dilation,  # type: ignore[arg-type]
                groups=self.offset_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('offset_generator', None)

        if self.modulated:
            self.mask_generator = nn.Conv2d(
                self.in_channels,
                self.kernel_size[0] * self.kernel_size[1] * self.mask_groups,
                kernel_size=self.kernel_size,  # type: ignore[arg-type]
                stride=self.stride,  # type: ignore[arg-type]
                padding=self.padding,
                dilation=self.dilation,  # type: ignore[arg-type]
                groups=self.mask_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('mask_generator', None)

        self.offset_activation = offset_activation
        self.mask_activation = mask_activation

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not hasattr(self, 'modulated'):
            return
        super().reset_parameters()
        if self.offset_generator is not None:
            init.zeros_(self.offset_generator.weight)
            if self.offset_generator.bias is not None:
                init.zeros_(self.offset_generator.bias)
        if self.mask_generator is not None:
            init.zeros_(self.mask_generator.weight)
            if self.mask_generator.bias is not None:
                init.zeros_(self.mask_generator.bias)

    def forward(self, input: Tensor) -> Tensor:
        if self.deformable and self.offset_generator is not None:
            offset = self.offset_generator(input)
            if self.offset_activation is not None:
                offset = self.offset_activation(offset)
        else:
            offset = None

        if self.modulated and self.mask_generator is not None:
            mask = self.mask_generator(input)
            if self.mask_activation is not None:
                mask = self.mask_activation(mask)
        else:
            mask = None

        return self._conv_forward(input, self.weight, offset, mask, self.bias)


# noinspection PyMethodOverriding
class PackedDeformConv3d(DeformConv3d):
    """
    Packed version of :class:`DeformConv3d`.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: _size_3_t,
                 stride: _size_3_t = 1,
                 padding: Union[str, _size_3_t] = 0,
                 dilation: _size_3_t = 1,
                 groups: int = 1,
                 offset_groups: int = 1,
                 mask_groups: int = 1,
                 bias: bool = True,
                 generator_bias: bool = False,
                 deformable: bool = True,
                 modulated: bool = False,
                 offset_activation: Optional[Union[str, Callable]] = None,
                 mask_activation: Optional[Union[str, Callable]] = 'sigmoid',
                 padding_mode: str = 'zeros',
                 device=None,
                 dtype=None) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            groups, bias, padding_mode, device, dtype)

        if in_channels % offset_groups != 0:
            raise ValueError('in_channels must be divisible by offset_groups')
        if out_channels % offset_groups != 0:
            raise ValueError('out_channels must be divisible by offset_groups')

        if in_channels % mask_groups != 0:
            raise ValueError('in_channels must be divisible by mask_groups')
        if out_channels % mask_groups != 0:
            raise ValueError('out_channels must be divisible by mask_groups')

        if isinstance(offset_activation, str):
            raise ValueError('currently, no activation is supported for offset')
        if isinstance(mask_activation, str):
            if mask_activation == 'sigmoid':
                mask_activation = MaskSigmoid(scale=2.)
            elif mask_activation == 'softmax':
                mask_activation = MaskSoftmax3d(self.kernel_size)  # type: ignore[arg-type]
            else:
                raise ValueError('only \"sigmoid\" and \"softmax\" activations are supported for mask')

        self.offset_groups = offset_groups
        self.mask_groups = mask_groups

        self.deformable = deformable
        self.modulated = modulated

        if self.deformable:
            self.offset_generator = nn.Conv3d(
                self.in_channels,
                3 * self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2] * self.offset_groups,
                kernel_size=self.kernel_size,  # type: ignore[arg-type]
                stride=self.stride,  # type: ignore[arg-type]
                padding=self.padding,
                dilation=self.dilation,  # type: ignore[arg-type]
                groups=self.offset_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('offset_generator', None)

        if self.modulated:
            self.mask_generator = nn.Conv3d(
                self.in_channels,
                self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2] * self.mask_groups,
                kernel_size=self.kernel_size,  # type: ignore[arg-type]
                stride=self.stride,  # type: ignore[arg-type]
                padding=self.padding,
                dilation=self.dilation,  # type: ignore[arg-type]
                groups=self.mask_groups,
                bias=generator_bias,
                device=device,
                dtype=dtype)
        else:
            self.register_module('mask_generator', None)

        self.offset_activation = offset_activation
        self.mask_activation = mask_activation

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if not hasattr(self, 'modulated'):
            return
        super().reset_parameters()
        if self.offset_generator is not None:
            init.zeros_(self.offset_generator.weight)
            if self.offset_generator.bias is not None:
                init.zeros_(self.offset_generator.bias)
        if self.mask_generator is not None:
            init.zeros_(self.mask_generator.weight)
            if self.mask_generator.bias is not None:
                init.zeros_(self.mask_generator.bias)

    def forward(self, input: Tensor) -> Tensor:
        if self.deformable and self.offset_generator is not None:
            offset = self.offset_generator(input)
            if self.offset_activation is not None:
                offset = self.offset_activation(offset)
        else:
            offset = None

        if self.modulated and self.mask_generator is not None:
            mask = self.mask_generator(input)
            if self.mask_activation is not None:
                mask = self.mask_activation(mask)
        else:
            mask = None

        return self._conv_forward(input, self.weight, offset, mask, self.bias)
